"""
  A library to generate gcode for milling PCB  
"""

# CVS $Id: ZetaTurtle.py,v 1.3 2012-05-05 08:41:12 zellini Exp $

import math
from copy import deepcopy
import sys,re,getopt,string
import xml.etree.ElementTree as etree
import matrix


class GC:
  """ one cent Graphic Context """
  # http://www.w3.org/TR/SVG/coords.html#TransformMatrixDefined
  def __init__ ( self):    
    self.default()
    self.clear()

  def default(self):
    self.a=1;self.b=0;self.c=0;self.d=1;self.e=0;self.f=0
    self.m =[ [self.a, self.c, self.e], [self.b, self.d, self.f], [0.0, 0.0, 1.0]]    

  def clear (self):
    self.vxy=[]

  def m2abcdef(self):
    self.a = self.m[0][0]; self.b=self.m[1] [0]; self.c=self.m[0][1]; self.d=self.m[1][1]; self.e=self.m[0][2]; self.f=self.m[1][2]
    
    
  def matrix (self, a, b, c, d, e, f):
    tm =[ [a, c, e], [b, d, f], [0.0, 0.0, 1.0]]
    self.m = matrix.mult(self.m, tm)
    self.m2abcdef()
    
    
  def translate (self,tx,ty):
    # or [1 0 0 1 tx ty], where tx and ty are the distances to translate coordinates in X and Y, respectively.
    tm = [[1, 0, tx], [0, 1, ty], [0, 0, 1]]
    self.m = matrix.mult(self.m, tm)
    self.m2abcdef()
    
  
  def scale (self,sx,sy):
    # or [sx 0 0 sy 0 0]. One unit in the X and Y directions in the new coordinate system equals sx and sy units in the previous coordinate
    tm = [[sx, 0, 0], [0, sy, 0], [0, 0, 1]]
    self.m = matrix.mult(self.m, tm)
    self.m2abcdef()


  def rotate (self,a):
    # rotation
    tm = [[math.cos(a),-math.sin(a) , 0], [math.sin(a), math.cos(a), 0], [0, 0, 1]]
    self.m = matrix.mult(self.m, tm)
    self.m2abcdef()

  def transform (self,xy):
    tm = [[xy[0]], [xy[1]], [1]]
    tmx = matrix.mult(self.m, tm)
    return (tmx[0][0], tmx[1][0])
  
  def transform1 (self,xy):
    xp = xy[0] * self.a + xy[1] * self.c + self.e
    yp = xy[0] * self.b + xy[1] * self.d + self.f
    return (xp,yp)
  
  def plot (self,xy):
    self.vxy.append (self.transform (xy))
    
  def arc (self, cx, cy, r, sa, da, dire):
    #print "arc center [%f,%f], radius %f startangle %f, delta %f, dir %s" % (cx, cy,r,  math.degrees(sa), math.degrees(da), dire)
    ac=0
    while ac<da:
      angle=sa+ac if dire else sa-ac
      ac=ac+.01
      x=math.cos(angle)*r+cx
      y=math.sin(angle)*r+cy
      self.plot((x, y))

  def arcq (self,cx,cy,rx,ry,sa,da,direction):
    ac=0
    while ac<da:
      angle=sa+ac if dir else sa-ac
      ac=ac+.01
      x=math.cos(angle)*rx+cx
      y=math.sin(angle)*ry+cy
      self.plot((x, y))
    

def cvf (n):
  return n * 25.4 / 90

def getCoeff(x):
  line = [1]
  for i in range(x):
    line = [0] + line
    for j in range(len(line)-1):
     line[j] += line[j+1]
  return line


def bez (delta, data):
  """ compute Bezier point """
  coeff = getCoeff(len(data)-1)
  exp = [(i,len(data)-(i+1)) for i in range(len(data))]
  t = 0
  points = []
  while t <= 1:
    x = 0
    y = 0
    for i in range(len(data)):
      x += coeff[i]*data[i][0]*(t**(exp[i][0]))*((1-t)**(exp[i][1]))
      y += coeff[i]*data[i][1]*(t**(exp[i][0]))*((1-t)**(exp[i][1])) 
    points.append((x,y))
    t += delta

  points.append ((data[-1][0],data[-1][1]))
  return points

def computeBezier (delta, data):
  inp=[]
  for d in data: inp.append ((d.x,d.y))
  points=[]
  pts = bez (delta,inp)
  for pt in pts:
    points.append(P(pt[0],pt[1]))        
  return points


def convertToFloats(aList,convert=True):
  "Convert number strings in list to floats (leave rest untouched)."

  for i in xrange(len(aList)):
    try:      
      aList[i] = cvf ( float(aList[i])) if convert else  float(aList[i])
    except ValueError:
      try:
          aList[i] = aList[i].encode("ASCII")
      except:
          pass

  return aList


def fixSvgPath(aList):
    """Normalise certain "abnormalities" in SVG paths.

    Basically, this reduces adjacent number values for h and v
    operators to the sum of these numbers and those for H and V
    operators to the last number only.

    Returns a slightly more compact list if such reductions
    were applied or a copy of the same list, otherwise.
    """

    # this could also modify the path to contain an op code
    # for each coord. tuple of a tuple sequence... 

    hPos, vPos, HPos, VPos, numPos = [], [], [], [], []
    for i in xrange(len(aList)):
        hPos.append(aList[i]=='h')
        vPos.append(aList[i]=='v')
        HPos.append(aList[i]=='H')
        VPos.append(aList[i]=='V')
        numPos.append(type(aList[i])==type(1.0))

    fixedList = []

    i = 0
    while i < len(aList):
        if hPos[i] + vPos[i] + HPos[i] + VPos[i] == 0:
            fixedList.append(aList[i])
        elif hPos[i] == 1 or vPos[i] == 1:
            fixedList.append(aList[i])
            sum = 0
            j = i+1
            while j < len(aList) and numPos[j] == 1:
                sum = sum + aList[j]
                j = j+1
            fixedList.append(sum)
            i = j-1
        elif HPos[i] == 1 or VPos[i] == 1:
            fixedList.append(aList[i])
            last = 0
            j = i+1
            while j < len(aList) and numPos[j] == 1:
                last = aList[j]
                j = j+1
            fixedList.append(last)
            i = j-1
        i = i+1

    return fixedList



def normaliseSvgPath(attr):
    # operator codes mapped to the minimum number of expected arguments 
    ops = {'A':7, 'a':7,
      'Q':4, 'q':4, 'T':2, 't':2, 'S':4, 's':4, 
      'M':2, 'L':2, 'm':2, 'l':2, 'H':1, 'V':1,  
      'h':1, 'v':1, 'C':6, 'c':6, 'Z':0, 'z':0}

    # do some preprocessing
    opKeys = ops.keys()
    a = attr
    a = a.replace(',', ' ')
    a = string.replace(a, 'e-', 'ee')
    a = string.replace(a, '-', ' -')
    a = string.replace(a, 'ee', 'e-')
    for op in opKeys:
        a = a.replace(op, " %s " % op)
    a = a.strip()
    a = a.split()
    #a = convertToFloats(a)
    a = fixSvgPath(a)

    # insert op codes for each argument of an op with multiple arguments
    res = []
    i = 0
    while i < len(a):
      el = a[i]
      if el in opKeys:
        lastel = el
        res.append(el)
        if el in ('z', 'Z'):
          res.append([])          
        else:
          res.append(a[i+1:i+1+ops[el]])
          i = i + ops[el]
        i=i+1 
      else:
        while i < len(a):
          if a[i] not in opKeys:
            el=lastel
            if lastel in ('m','l'): el= 'l'
            res.append(el)
            res.append(a[i:i+ops[el]])
            i = i + ops[el]
          else:
            break

    return res

def normalize(string=None):
    """Normalize a string corresponding to an array of various vaues."""
    string = string.replace("-", " -")
    string = string.replace(",", " ")

    while "  " in string:
        string = string.replace("  ", " ")

    string = string.replace("e -", "e-")

    return string


def dotransform (transform):
  global width,height,gc
  transformations = transform.split(")")
  for transformation in transformations:
      for ttype in (
          "scale", "translate", "matrix", "rotate", "skewX",
          "skewY"):
          if ttype in transformation:
              transformation = transformation.replace(ttype, "")
              transformation = transformation.replace("(", "")
              transformation = normalize(transformation).strip() + " "
              values = []
              while transformation:
                value, transformation =  transformation.split(" ", 1)
                  # TODO: manage the x/y sizes here
                values.append(value)      
              if ttype=='translate':
                gc.translate (cvf (  float(values[0]))   , -cvf(float(values[1])  ))
              elif ttype=='matrix':
                a, b, c, d = convertToFloats (values[:4], False)
                e, f = convertToFloats (values[4:])
                f=height-(d * height+f)
                e=e+c*height
                gc.matrix (a, -b, -c, d, e, f)
                pass
              else:
                pass


def readSVG(s):
  global width,height,gc

  tree = etree.parse (s)
  svg = tree.getroot()
  width = cvf ( float(svg.attrib['width']))
  height = cvf (float( svg.attrib['height']))

  outPaths=[]

  entries = tree.findall('{http://www.w3.org/2000/svg}g')
  for entry in entries:
    label = entry.attrib["{http://www.inkscape.org/namespaces/inkscape}label"]
    if label=='path':
      # check for transform in entry
      firsttr=False
      if entry.attrib.has_key("transform"):
        firsttr = entry.attrib["transform"]
        #print "*****Transform: %s" % (firsttr)


      childs = entry.getchildren()
      for path in childs:
        """ do a child ( path, rect, etc )  """        
        if path.get('gc'):
          gc = path.get('gc')
        else:        
          gc=GC()   # element GC

        if firsttr:
          dotransform(firsttr)          
        transform=None
        if path.attrib.has_key("transform"):
          transform = path.attrib["transform"]
          #print "Transform: %s" % (transform)
          dotransform (transform)
        
        # print path.tag
        if path.tag=='{http://www.w3.org/2000/svg}g':
          """ 2 cent recursion simulation """
          for ch in path.getchildren():
            if transform: 
              ch.set('gc',gc)
            childs.append (ch)

        if path.tag=='{http://www.w3.org/2000/svg}rect':
          """ do a rect """
          hrect = path.attrib["height"]
          wrect = path.attrib["width"]
          xrect = path.attrib["x"]
          yrect = path.attrib["y"]
          rx = path.attrib["rx"] if path.attrib.has_key("rx") else 0
          ry = path.attrib["ry"] if path.attrib.has_key("ry") else 0
          #print "Rect x %s, y %s, width %s, height %s, rx %s, ry %s" % (xrect,yrect,wrect,hrect,rx,ry)
          hrect,wrect,xrect,yrect,rx,ry = convertToFloats([hrect,wrect,xrect,yrect,rx,ry])
          yrect = height-yrect
          #print "Rect x %s, y %s, width %s, height %s, rx %s, ry %s" % (xrect,yrect,wrect,hrect,rx,ry)
          XY=[]
          if rx==0 or ry==0:
            XY.append (gc.transform ((xrect,yrect)))
            XY.append (gc.transform ((xrect+wrect,yrect)))
            XY.append (gc.transform ((xrect+wrect,yrect-hrect)))
            XY.append (gc.transform ((xrect,yrect-hrect)))
            XY.append (gc.transform ((xrect,yrect)))
          else:
            XY.append (gc.transform ((xrect+rx,yrect)))
            XY.append (gc.transform ((xrect+wrect-rx,yrect)))
            gc.clear()
            gc.arcq   (xrect+wrect-rx, yrect-ry, -rx,ry,math.pi/2,math.pi/2,False)
            for pt in gc.vxy:
              XY.append(pt)
            XY.append (gc.transform ((xrect+wrect,yrect-hrect+ry)))
            gc.clear()
            gc.arcq   (xrect+wrect-rx, yrect-hrect+ry, rx,-ry,0,math.pi/2,False)
            for pt in gc.vxy:
              XY.append(pt)
            XY.append (gc.transform ((xrect+rx,yrect-hrect)))
            gc.clear()
            gc.arcq   (xrect+rx, yrect-hrect+ry, -rx,ry,-math.pi/2,math.pi/2,False)
            for pt in gc.vxy:
              XY.append(pt)
            XY.append (gc.transform ((xrect,yrect-ry)))
            gc.clear()
            gc.arcq   (xrect+rx, yrect-ry, rx,-ry,math.pi,math.pi/2,False)
            for pt in gc.vxy:
              XY.append(pt)

            
          outPaths.append(XY)



        if path.tag=='{http://www.w3.org/2000/svg}path':
          """ do a path"""
          XY=[]
          px=0;py=0
          dd = path.attrib["d"]
          normPath = normaliseSvgPath(dd)
          for i in xrange(0, len(normPath), 2):
            op, nums = normPath[i:i+2]
            relative = op.islower() # True =relative
            op=op.lower()
            if op =='m':
              x,y = convertToFloats(nums)
              y=height-y
              # start of path
              px=x;py=y
              sx=x;sy=y
              XY.append (gc.transform((px,py)))
          
            elif op=='l':
              x,y = convertToFloats(nums)
              y = -y if relative else height-y
              px = px+x if relative else x
              py = py+y if relative else y
              #print "line to %f,%f" % (px,py)
              XY.append (gc.transform((px,py)))
            elif op=='c':
              """ cubic bezier """
              x1,y1,x2,y2,x,y = convertToFloats(nums)
              y1 = -y1 if relative else height-y1
              y2 = -y2 if relative else height-y2
              y =  -y if relative else height-y
              spx = px; spy = py
              if relative: x1+=spx
              if relative: x2+=spx
              if relative: y1+=spy
              if relative: y2+=spy
              px = px+x if relative else x
              py = py+y if relative else y
              #print "bezier x1 %f,y1 %f,x2 %f,y2 %f,x %f,y %f" % (x1,y1,x2,y2,x,y)
              #print "bezier from %f,%f from %f,%f through %f,%f and %f,%f" % (spx,spy,px,py,x1,y1,x2,y2)
              pts = bez (0.01,[(spx,spy),(x1,y1),(x2,y2),(px,py)])
              for pt in pts:
                XY.append(gc.transform(pt))

              #XY.append((px,py))

            elif op=='a':
              """ elliptical arc (rx ry x-axis-rotation large-arc-flag sweep-flag x y) """
              # http://www.w3.org/TR/SVG/implnote.html#ArcImplementationNotes
              x1 = px; y1 = py
              rx,ry,xrot,large,sweep,x2,y2 = convertToFloats(nums,False)
              #print "Arc radii %s,%s x-axis-rotation %s large-arc %s sweep-flag %s x %s y %s" % (rx,ry,xrot,large,sweep,x2,y2)
              rx,ry,x2,y2 = convertToFloats ([rx,ry,x2,y2])
              # convert x2, y2 from relative to absolute
              large = bool (large); sweep = bool (sweep)
              y2 = -y2 if relative else height-y2
              if relative: x2+=px
              if relative: y2+=py
              px=x2;py=y2
              
              # TODO: check parameter 

              phi = xrot*math.pi/180.0
              #print "(x1=%f,y1=%f) Arc radii %f,%f x-axis-rotation %f large-arc %s sweep-flag %s x %f y %f" % (x1,y1,rx,ry,xrot,large,sweep,x2,y2)
              # x1',y1'
              x1p = math.cos(phi) * ((x1-x2) / 2) + math.sin(phi) * ((y1-y2) / 2) 
              y1p = -math.sin(phi)* ((x1-x2) / 2) + math.cos(phi) * ((y1-y2) / 2) 
              # adjust radii
              l = math.pow(x1p,2)/math.pow(rx,2)+math.pow(y1p,2)/math.pow(ry,2);
              if l>1:
                rx *= math.sqrt(l);
                ry *= math.sqrt(l);

              try:
    					  s = (-1 if large == sweep else 1) * math.sqrt(((math.pow(rx,2)*math.pow(ry,2))-(math.pow(rx,2)*math.pow(y1p,2))-(math.pow(ry,2)*math.pow(x1p,2))) /(math.pow(rx,2)*math.pow(y1p,2)+math.pow(ry,2)*math.pow(x1p,2)));
              except:
                s=0
              cxp = s *  rx * y1p / ry
              cyp = s * -ry * x1p / rx
              # cx, cy
              cx = (x1 + x2) / 2.0 +    math.cos(phi) * cxp - math.sin(phi) * cyp
              cy = (y1 + y2) / 2.0 +    math.sin(phi) * cxp + math.cos(phi) * cyp

              m = lambda v: math.sqrt(math.pow(v[0],2) + math.pow(v[1],2))
              r = lambda u, v: (u[0]*v[0]+u[1]*v[1]) / (m(u)*m(v)) 
              # angle between two vectors
              a = lambda u, v: (-1 if u[0]*v[1] < u[1]*v[0] else 1) * math.acos(r(u,v))
              # initial angle
              a1 = a((-1,0), ((x1p-cxp)/rx,(y1p-cyp)/ry));
              # angle delta
              u = ((x1p-cxp)/rx,(y1p-cyp)/ry);
              v = ((-x1p-cxp)/rx,(-y1p-cyp)/ry);
              ad = a(u, v);
              if (r(u,v) <= -1): ad = math.pi
              if (r(u,v) >= 1): ad = 0

              if (not sweep  and ad > 0): ad = ad - 2 * math.pi;
              if (sweep and ad < 0): ad = ad + 2 * math.pi;

              r = rx if rx > ry else ry
              ssx = 1 if rx > ry else rx / ry
              ssy = ry/rx if rx > ry else 1
              
              a1=a1+ad
              #print "a1 %f ad %f" % (math.degrees(a1),math.degrees(ad))
              xxp = x1-math.cos(a1) * rx
              yyp = y1-math.sin(a1) * ry
              #print "xxp %f, yyp %f" % (xxp, yyp)
              
              gc.translate(xxp, yyp)
              gc.rotate (phi)
              gc.scale(ssx, ssy)
              gc.clear()
              gc.arc (0, 0, r, a1, ad, not sweep)
              
              gc.scale (1/ssx, 1/ssy)
              gc.rotate(-phi)
              gc.translate(-xxp, -yyp)
              
              for pt in gc.vxy:
                XY.append(pt)

            elif op=='z':              
              XY.append (gc.transform((sx,sy)))
            else:
              pass
          if len(XY)>1: outPaths.append(XY)

  return outPaths            


class Scale:
  def __init__(self,x,y,z):
    """ a scale """
    self.x=float(x)
    self.y=float(y)
    self.z=float(z)
    if self.x==self.y:
      self.XY=self.x


  def __str__(self):
    return "Scale %f,%f,%f" % (self.x,self.y,self.z)


class Offset:
  def __init__(self,x,y,z):
    """ offset """
    self.x=float(x)
    self.y=float(y)
    self.z=float(z)

  def __str__(self):
    return "Offset %f,%f,%f" % (self.x,self.y,self.z)

class P:
  def __init__(self,x=None,y=None,z=None,r=False):
    """ A point in  space """
    self.x= float(x) if x!=None else None
    self.y= float(y) if y!=None else None
    self.z= float(z) if z!=None else None
    self.relative= r

  def __str__(self):
    s="P "
    if self.x!=None: s+= ("%f " % self.x)
    if self.y!=None: s+= ("%f " % self.y)
    if self.z!=None: s+= ("%f " % self.z)
    return s

class PR (P):
  def __init__(self,x=None,y=None,z=None):
    """ A relative point in  space """
    P.__init__(self,x,y,z,True)


class Drill:

  def __init__(self,p,offset,scale):
    """ current position """
    """ units not scaled: scale performed when commands are issued """
    self.p=p
    self.offset=offset
    self.scale=scale
    self.debugLevel=0
    self.sock=sys.stdout

  def emit (self,s):
    self.sock.write ("%s\n" % (s))
    #sys.stdout.write ("%s\n" % (s))


  def sin(self,x,l=1):
    return math.sin(math.radians(x))*l

  def cos(self,x,l=1):
    return math.cos(math.radians(x))*l
 

  def debug (self,msg):
    if self.debugLevel:
      print msg

  def getActualDrillPosition(self):
    return self.getDrillPositionFor(self.p)

  def getDrillPositionFor(self,p):
    """ return true cnc Point """
    if p.relative:
      x = p.x+self.p.x if p.x!=None else self.p.x
      y = p.y+self.p.y if p.y!=None else self.p.y
      z = p.z+self.p.z if p.z!=None else self.p.z
    else:
      x = p.x if p.x!=None else self.p.x
      y = p.y if p.y!=None else self.p.y
      z = p.z if p.z!=None else self.p.z

    return P( ( x + self.offset.x) * self.scale.x,
              ( y + self.offset.y) * self.scale.y,
              ( z + self.offset.z) * self.scale.z )

  def setCurrentP(self,p):
    """ set current position """
    if p.relative:
      if p.x!= None: self.p.x+= p.x
      if p.y!= None: self.p.y+= p.y
      if p.z!= None: self.p.z+= p.z
    else:
      if p.x!= None: self.p.x = p.x
      if p.y!= None: self.p.y = p.y
      if p.z!= None: self.p.z = p.z

    
  def goto(self,p,g="G1"):
    """ goto point """
    
    arc = re.match ("^G[23]",g)
    c = self.getActualDrillPosition()
    n = self.getDrillPositionFor (p)
    cmd=""
    sp=""
    if (c.x != n.x or arc ):
      cmd=cmd+ ( "X%f" % (n.x)) 
      sp=" "
    if (c.y != n.y or arc ):
      cmd=cmd + sp + ( "Y%f" % (n.y)) 
      sp=" "
    if (c.z != n.z):
      cmd=cmd + sp + ( "Z%f" % (n.z)) 

    if len (cmd) or arc:
      self.emit ("%s %s" % (g,cmd))  
      self.setCurrentP(p)   

  def arc(self,p,r,mode):
    """ arc to point, clock wise, radius """
    self.setCurrentP(p)
    c = self.getActualDrillPosition()
    self.emit ("%s X%lf Y%lf R%lf" % (mode,c.x,c.y,r*self.scale.XY))

    
class PCBDrill (Drill):

  def __init__(self,p,offset,scale,safeHeight):
    Drill.__init__(self,p,offset,scale)
    self.down=True
    self.safeHeight=safeHeight

  def safeUp (self):
    """ Rise drill, for PCB """
    Drill.goto (self,P(z=self.safeHeight),"G0")
    self.down=False

  def pungle (self):
    """ Rise drill, for PCB """
    Drill.goto (self,P(z=0),"G1 F40")
    self.down=True

  def goto (self,p):
    Drill.goto (self,p, "G1" if self.down else "G0")


  def arc (self,p,radius=1,mode="cw"):
    """ arc """
    Drill.arc( self,p , radius, "G2" if mode=="cw" else "G3" )


  """ high level commands """

  def circle (self,p,radius=1,mode="cw"):
    """ drill a circle, given center, mode and radius """
    self.safeUp()
    self.goto (p)
    self.goto (PR(y=-radius))
    self.pungle()
    pos = deepcopy (self.p)
    """ compute relative center coordinates (I always 0, only J) """
    Drill.goto(self,pos,g="G2 J%f" % (radius*self.scale.XY) )
    self.safeUp()
    Drill.goto (self,PR(y=radius))

  def poly (self,points):
    """ do a polygon """
    self.safeUp()
    first=True
    for point in points:
      if first:
        self.goto (point)
        self.pungle()
        first=False
      else:
        self.goto (point)

    self.safeUp()

  def bezier (self,delta,points):
    """ first point relative/absolute to machine """
    """ all other points relative to first point """
    first=True
    rp = []
    for point in points:
      if first:
        first=False
        if point.relative:
          x= self.p.x + (point.x if point.x!=None else 0)
          y= self.p.y + (point.y if point.y!=None else 0)
        else:
          x = (point.x if point.x!=None else self.p.x)
          y = (point.y if point.y!=None else self.p.y)
        xs = x
        ys = y
      else:
        if point.relative:
          x = xs + (point.x if point.x!=None else 0)
          y = ys + (point.y if point.y!=None else 0)
        else:
          x = (point.x if point.x!=None else xs)
          y = (point.y if point.y!=None else ys)


      rp.append(P(x,y))

    p = computeBezier (delta, rp)
    self.poly (p)



  def square (self,start,w,h):
    """ do a square given start, width and height """
    self.poly ([start, PR (x=w),PR(y=h), PR(x=-w),PR(y=-h)])

  def squareRounded (self,start,w,h,radius):
    """ do a square given start, width and height """
    self.safeUp()
    self.goto(start)
    self.goto(PR(x=radius))
    self.pungle()
    self.goto (PR(x=w-radius*2))
    self.arc (PR(x=radius,y=radius),radius=radius,mode="ccw")
    self.goto (PR(y=h-radius*2))
    self.arc (PR(y=radius,x=-radius),radius=radius, mode="ccw")
    self.goto (PR(x=-w+radius*2))
    self.arc (PR(y=-radius,x=-radius),radius=radius,mode="ccw")
    self.goto (PR(y=-h+radius*2))
    self.arc (PR(y=-radius,x=radius),radius=radius,mode="ccw")

    self.safeUp()
    self.goto(PR(x=-radius))

  def svg (self,s,section='path'):
    paths = readSVG(s)  # TODO: coords are in mm, resize to scale     
    for path in paths:
      XY=[]
      for x,y in path:
        XY.append (P(x,y))
      self.poly (XY)


if __name__ == "__main__":
  """ test stuffs """
  d = PCBDrill (P(0,0,0),Offset(0,0,0),Scale(1,1,1),3)


  d.sock = open ('pippo.ngc','w')
  d.emit ("G21\n")
  d.emit ("G94\n")
  d.svg ("trap130.svg")
  d.emit ("M2")
  sys.exit(1)


  d.circle (PR(y=1,x=1),radius=1)
  d.circle (PR(x=2,y=2),radius=1.5)
  d.poly ([P(x=2,y=0), PR(y=1), PR(x=1)])
  d.square (P(x=10,y=10),4,2)
  d.squareRounded (P(x=7,y=4),4,4,0.5)

  d.poly ([P(x=0,y=10), PR(x=4)])

  d.bezier (0.0005,[P(x=10,y=10),PR(y=.7,x=.3),  PR(x=1,y=.5),  PR(x=2-.3,y=0.3),PR(x=2,y=1)])

  #d.poly ([P(), PR(x=4)])

  d.emit ("M2")





